/*
 libualloc stacks example - this example code is placed in the public domain
 by Tibor 'Igor2' Palinkas in 2020

 This example implements tac(1) for real numbers: it reads a whitespace 
 separated list of numbers on stdin and prints them in reverse order,
 one number per line.

 Rationale for using stacks, the fixed size stack allocator
 ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

 Length of input is unknown and all data needs to be read before output
 can be generated, which means dynamic allocation can not be avoided.
 Using a linked list costs one allocation per number (also an extra
 pointer per number); using a growing vector with realloc() costs no memory
 overhead but a lot of memcpy().

 A more efficient implementation is using a list of arrays, each array
 holding N numbers, arrays put on a singly linked list. This is exactly what
 libualloc stacks implements: it allocates memory in bigger chunks, called
 pages, stores elements tightly packed within the page and links the pages
 together.

 Stacks is optimal for this specific allocation scheme: allocate sequentially,
 free them in reverse order.

 It is also optimal if allocated elements are used as nodes to build a tree
 or linked lists or other complex data structures, loading a data file or
 parsing some grammar. In that case the caller will not ever use the top()
 or free() function but will allocate all nodes sequentially and free the
 resulting stacks at once, when the data structure is discarded, using
 a single clean() call.

 Page size is set to 4096 bytes; on x86_64 stacks will use 24 bytes per
 page for storing administrative information which leaves 4072 bytes
 for storing numbers. A double is usually 64 bits long so a page can
 effectively store 509 doubles. Thus the overhead of page allocation is
 about 0.5%.
*/

#include <stdio.h>
#include <libualloc/libualloc.h>
#include <libualloc/stacks_api.h>

int main(int argc, char *argv[])
{
	uall_sysalloc_t sys = {0};
	uall_stacks_t st = {0};
	double tmp, *d;

	/* initialize the low level allocator: use libc's malloc/free */
	sys.alloc     = uall_stdlib_alloc;
	sys.free      = uall_stdlib_free;
	sys.page_size = 4096; /* how much memory to allocate to grow the stack */

	/* initialize the stack allocator */
	st.sys = &sys;
	st.elem_size = sizeof(double);

	/* load input */
	while(scanf("%lf", &tmp) == 1) {
		d = uall_stacks_alloc(&st);
		if (d == NULL)
			return 1; /* in case sys.alloc a.k.a. malloc() fails on page allocation */
		*d = tmp;
	}

	/* print result */
	for(;;) {
		d = uall_stacks_top(&st);
		if (d == NULL)
			break;
		printf("%f\n", *d);
		uall_stacks_free(&st);
	}

	return 0;
}
