/*

libualloc - microlib for memory allocation with various strategies

Copyright (c) 2020 Tibor 'Igor2' Palinkas
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. Neither the name of the Author nor the names of contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

Source code: svn://svn.repo.hu/libualloc/trunk
Contact the author: http://igor2.repo.hu/contact.html

*/

#ifdef TESTER_SLAB_AP
#	include <libualloc/slabap_api.h>
#	include <libualloc/slabap_impl.h>
#	define MAIN_NAME main_slabap
#	define DRAW draw_slabap
#	define PAGE_T uall_slabap_page_t
#	define SLAB_T uall_slabap_t
#	define SLAB2PAGE(ctx, slab) uall_slabap2page(ctx, slab)
#	define SLAB_ALLOC uall_slabap_alloc
#	define SLAB_FREE uall_slabap_free
#	define SLABS_PER_PAGE UALL_SLABAPS_PER_PAGE
#	define SLAB_REAL_SIZE UALL_SLABAP_REAL_SIZE
#	define SLAB_PTR2SLAB UALL_SLABAP_PTR2SLAB
#	define SLABS_T uall_slabaps_t
#	define SLAB_CLEAN uall_slabap_clean

void *handalign(void *ptr_, long size)
{
	size_t ptr = (size_t)ptr_;

	if ((size & (size - 1)) == 0) {
		size_t mask = ~(size - 1);
		return (void *)((ptr & mask) + size);
	}
	return (void *)((ptr / size * size) + size);
}

void *uall_stdlib_alloc_handalign(uall_sysalloc_t *ctx, long size_)
{
	size_t size = ctx->page_size;
	void *tmp = malloc(size*2+sizeof(void *));
	void *res = handalign(tmp, size);
	void **mark = (void **)((char *)res + size);

	*mark = tmp;

/*	printf("ALLOC: %p -> %p\n", tmp, res);*/

	return res;
}

void uall_stdlib_free_handalign(uall_sysalloc_t *ctx, void *block)
{
	void **mark = (void **)((char *)block + ctx->page_size);
	void *tmp = *mark;

/*	printf("ALLOC: %p <- %p\n", tmp, block);*/

	return free(tmp);
}

#else
#	include <libualloc/slab_api.h>
#	include <libualloc/slab_impl.h>
#	define MAIN_NAME main_slab
#	define DRAW draw_slab
#	define PAGE_T uall_slab_page_t
#	define SLAB_T uall_slab_t
#	define SLAB2PAGE(ctx, slab) slab->page
#	define SLAB_ALLOC uall_slab_alloc
#	define SLAB_FREE uall_slab_free
#	define SLABS_PER_PAGE UALL_SLABS_PER_PAGE
#	define SLAB_REAL_SIZE UALL_SLAB_REAL_SIZE
#	define SLAB_PTR2SLAB UALL_SLAB_PTR2SLAB
#	define SLABS_T uall_slabs_t
#	define SLAB_CLEAN uall_slab_clean
#endif

static void DRAW(SLABS_T *sl, pool_t *pool)
{
	char *line;
	long p, n, len, num_pages = 0;
	PAGE_T *page;
	SLAB_T *s;

	if (sl->pages == NULL) {
		printf("<empty, no page allocated>\n");
		return;
	}

	for(page = sl->pages; page != NULL; page = page->next)
		num_pages++;

	len = num_pages * (SLABS_PER_PAGE(sl)+1);
	line = malloc(len);

	/* render pages and slabs, all unclassified */
	memset(line, '?', len-1);
	line[len] = '\0';
	for(n = SLABS_PER_PAGE(sl); n < len; n += SLABS_PER_PAGE(sl)+1)
		line[n] = ' ';

	/* mark free slabs */
	for(s = sl->free_slabs; s != NULL; s = s->next) {
		for(n = 0, page = sl->pages; page != NULL; n += SLABS_PER_PAGE(sl)+1, page = page->next) {
			if (page == SLAB2PAGE(sl, s)) {
				int offs = ((char *)s - page->data) / SLAB_REAL_SIZE(sl);
				line[n+offs] = '.';
				break;
			}
		}
	}

	/* mark allocated slabs */
	for(p = 0; p < pool->used; p++) {
		s = SLAB_PTR2SLAB(pool->arr[p]);
		for(n = 0, page = sl->pages; page != NULL; n += SLABS_PER_PAGE(sl)+1, page = page->next) {
			if (page == SLAB2PAGE(sl, s)) {
				int offs = ((char *)s - page->data) / SLAB_REAL_SIZE(sl);
				line[n+offs] = 'a';
				break;
			}
		}
	}

	printf("%s\n", line);

	free(line);
}

int MAIN_NAME(int argc, char *argv[])
{
	uall_sysalloc_t sys;
	SLABS_T sl = {0};
	pool_t pool = {0};
	char *cmd, *arg;
	void *s;
	int its = 128, n, trace = 0, alloc_cnt = 0, free_cnt = 0;

#ifdef TESTER_SLAB_AP
	sys.alloc = uall_stdlib_alloc_handalign;
	sys.free = uall_stdlib_free_handalign;
#else
	sys.alloc = uall_stdlib_alloc;
	sys.free = uall_stdlib_free;
#endif
	sys.page_size = 128;

	for(n = 1; n < argc; n++) {
		cmd = argv[n]; arg = argv[n+1];
		if (cmd == NULL) continue;
		while(*cmd == '-') cmd++;
		switch(*cmd) {
			case 's': rnd_seed = atoi(arg); n++; break;
			case 'i': its = atoi(arg); n++; break;
			case 'p': sys.page_size = atoi(arg); n++; break;
			case 't': trace = 1; break;
			default:
				fprintf(stderr, "Unrecognized argument '%s'\n", argv[n]);
				return -1;
		}
	}


	sl.sys = &sys;
	sl.slab_size = 5;

	/* randomly allocate and free */
	for(n = 0; n < its; n++) {
		if (rnd() % 2) {
			s = SLAB_ALLOC(&sl);
			assert(s != NULL);
			pool_push(&pool, s);
			alloc_cnt++;
			if (trace)
				printf("alloc: %p\n", s);
		}
		else {
			s = pool_pop_rand(&pool);
			if (s != NULL) {
				if (trace)
					printf("free:  %p\n", s);
				SLAB_FREE(&sl, s);
				free_cnt++;
			}
		}
	}

	printf("*** Alloced: %d freed: %d\n", alloc_cnt, free_cnt);
	DRAW(&sl, &pool);

	/* all slabs that are not free'd are currently allocated and stored in the pool */
	assert((alloc_cnt - free_cnt) == pool.used);

	/* free all remaining slabs */
	SLAB_CLEAN(&sl);

	return 0;
}


#undef MAIN_NAME
#undef DRAW
#undef PAGE_T
#undef SLAB_T
#undef SLAB2PAGE
#undef SLAB_ALLOC
#undef SLAB_FREE
#undef SLABS_PER_PAGE
#undef SLAB_REAL_SIZE
#undef SLAB_PTR2SLAB
#undef SLABS_T
#undef SLAB_CLEAN

